﻿Import()


function Import()
{
    app.beginUndoGroup( "Callipeg to After Effects" );
    
    var fileURL = File.openDialog("Select data file:");
    if( fileURL != null && fileURL.name == "data" ) 
    {
        var proj = app.project;
        if (!proj) proj = app.newProject();
        
        var fileString = String(fileURL)
        var folderURL = Folder(fileString.substring(0, Math.max(fileString.lastIndexOf("/"), fileString.lastIndexOf("\\"))))  
        
        if (fileURL) 
        {
                var file = File (fileURL)
                file.open ("r")
                var text = file.readln();
                file.close()
        
                // Just parse JSON file, then you get a dictionnary easy to read
                jsonObj = JSON.parse( text );
                var movieLength = GetMovieLength( jsonObj )
                var movieFPS = jsonObj["fps"]
    
                var globalFolder = app.project.items.addFolder("Callipeg")
                var mainFolder = app.project.items.addFolder(jsonObj["name"])
                mainFolder.parentFolder = globalFolder;

                var compsFolder = app.project.items.addFolder("Comps")
                compsFolder.parentFolder = mainFolder;
                
                var artworkFolder = app.project.items.addFolder("Artwork")
                artworkFolder.parentFolder = mainFolder;

                var mainComp = proj.items.addComp( "_"+jsonObj["name"], jsonObj["width"], jsonObj["height"], 1, movieLength/movieFPS, movieFPS );
                mainComp.parentFolder = globalFolder;
    
                // var transformFolder = null
    
                // Read all layers
                var totalLayerCount = jsonObj["layers"].length;
                for( layerIndex=0; layerIndex < totalLayerCount; layerIndex++ ) 
                {
                    var layerInfoNode = jsonObj["layers"][totalLayerCount - layerIndex - 1]; // Reversed order
            
                    // Names in AE can't have special caracters, so we remove  em all
                    var layerNameAlphaNum = layerInfoNode["layer"]["name"].replace( /[^A-Za-z0-9]/g, "_" );
            
                    if( layerInfoNode["type"] == "Drawing" )
                    {
                        var layerComp = proj.items.addComp( layerNameAlphaNum, jsonObj["width"], jsonObj["height"], 1, movieLength/movieFPS, movieFPS );
                        layerComp.parentFolder = compsFolder;
                        precomplayer = mainComp.layers.add(layerComp);
                        precomplayer.property( "opacity" ).setValue( layerInfoNode["layer"]["opacity"] * 100 );
                        precomplayer.enabled = layerInfoNode["layer"]["visibility"];
                        
                        var layerFolder = app.project.items.addFolder( layerNameAlphaNum )
                        layerFolder.parentFolder = artworkFolder
                        for( instanceIndex=0; instanceIndex < layerInfoNode["layer"]["instances"].length; instanceIndex++ )
                        {
                            var instanceNode = layerInfoNode["layer"]["instances"][instanceIndex]
                            var instanceNameAlphaNum = instanceNode["instanceID"].replace( /[^A-Za-z0-9]/g, "_" )
                            instanceStartFrame = instanceNode["startingFrame"]
                            var instanceEndFrame = instanceNode["endingFrame"]
    
                            var fileURL = decodeURI( String(folderURL) + "/" + instanceNode["instanceID"] + ".png" )
                            var importOptions = new ImportOptions( fileURL )
                            var importedImage = app.project.importFile( importOptions )
                            importedImage.parentFolder = layerFolder
    
                            var instanceLayer = layerComp.layers.add( importedImage )
                            instanceLayer.inPoint = instanceStartFrame/movieFPS;
                            instanceLayer.outPoint = instanceEndFrame/movieFPS;
                            instanceLayer.name = instanceNameAlphaNum
                        }
                    }
                    else if ( layerInfoNode["type"] == "Transformation" )
                    {
                        // if ( transformFolder == null )
                        // {
                        //     transformFolder = app.project.items.addFolder("NullObjects")
                        //     transformFolder.parentFolder = mainFolder
                        // }
    
                        var transfoNullObject = mainComp.layers.addNull( movieLength )
                        transfoNullObject.name = layerNameAlphaNum
                        transfoNullObject.property("Position").dimensionsSeparated = true
                        transfoNullObject.enabled = layerInfoNode["layer"]["visibility"];
    
                        // Pivot setup
                        var pivotWeights = layerInfoNode["layer"]["pivotWeight"];
                        var pivotPosPixel = [ jsonObj["width"] * pivotWeights[0], jsonObj["height"] * pivotWeights[1] ]
                        transfoNullObject.transform.xPosition.setValue( pivotPosPixel[0] )
                        transfoNullObject.transform.yPosition.setValue( pivotPosPixel[1] )
    
                        // Parent layer here to children, because changing properties will move children only if linked
                        // If you set values up and then link, children layer will still be at original position, because not moved by parent as you set it up
                        // Connect layers
                        var childrenIndices = layerInfoNode["layer"]["childIndices"];
                        for ( childIndex=0; childIndex < childrenIndices.length; childIndex++ )
                        {
                            var theLayerIndex = childrenIndices[childIndex]
                            var theLayerName = jsonObj["layers"][ theLayerIndex ]["layer"]["name"].replace( /[^A-Za-z0-9]/g, "_" );
                            mainComp.layers.byName( theLayerName ).parent = transfoNullObject
                        }

                        var curves = layerInfoNode["layer"]["curves"];
                        BuildCurves( curves, transfoNullObject, movieFPS, pivotPosPixel[0], pivotPosPixel[1] )

                    }
                }
                
                
                for( itemIndex=1; itemIndex <= app.project.items.length; itemIndex++ ) 
                {
                    var item = app.project.items[ itemIndex ]
                    if( item.name == "Solids" )
                    {
                        item.parentFolder = globalFolder
                    }
                }
        }
    }
    
    app.endUndoGroup();
}


function GetMovieLength( jsonObj )
{
    var minFrame = 999999
    var maxFrame = -1
    var totalLayerCount = jsonObj["layers"].length;
    for( layerIndex=0; layerIndex < totalLayerCount; layerIndex++ ) 
    {
        var layerInfoNode = jsonObj["layers"][totalLayerCount - layerIndex - 1]; // Reversed order

        if( layerInfoNode["type"] == "Drawing" )
        {
            for( instanceIndex=0; instanceIndex < layerInfoNode["layer"]["instances"].length; instanceIndex++ )
            {
                var instanceNode = layerInfoNode["layer"]["instances"][instanceIndex];
                instanceStartFrame = instanceNode["startingFrame"];
                var instanceEndFrame = instanceNode["endingFrame"];
                
                if (instanceStartFrame < minFrame ) minFrame = instanceStartFrame
                if (instanceEndFrame > maxFrame ) maxFrame = instanceEndFrame
            }	
        }
    }

    return  maxFrame - minFrame;
}



function BuildCurves( curvesJSON, transfoNullObject, framerate, pivotX, pivotY )
{
    // Position X
    if ( curvesJSON[0]["keys"].length > 0 )
    {
        for( keyIndex=0; keyIndex < curvesJSON[0]["keys"].length; keyIndex++ )
        {
            var keyframeJSON = curvesJSON[0]["keys"][keyIndex]
            var lTan = keyframeJSON["lTangent"]
            var rTan = keyframeJSON["rTangent"]
            var posX = keyframeJSON["time"]
            var posY = pivotX + keyframeJSON["value"]

            BuildKeyFrame( transfoNullObject.transform.xPosition, framerate, curvesJSON[0]["keys"], keyIndex, posX, posY, lTan, rTan, pivotX, pivotY )
        }
    }

    // Position Y
    if ( curvesJSON[1]["keys"].length > 0 )
    {
        for( keyIndex=0; keyIndex < curvesJSON[1]["keys"].length; keyIndex++ )
        {
            var keyframeJSON = curvesJSON[1]["keys"][keyIndex]
            var lTan = keyframeJSON["lTangent"]
            var rTan = keyframeJSON["rTangent"]
            var posX = keyframeJSON["time"]
            var posY = pivotY + keyframeJSON["value"]

            BuildKeyFrame( transfoNullObject.transform.yPosition, framerate, curvesJSON[1]["keys"], keyIndex, posX, posY, lTan, rTan, pivotX, pivotY )
        }
    }

    // Scale X
    var scaleValues = BuildScale(curvesJSON[2]["keys"], curvesJSON[3]["keys"] )
    if( scaleValues != null )
    {
        for( keyIndex=0; keyIndex<scaleValues[0].length; keyIndex++ )
        {
            var xKeyframe = scaleValues[0][keyIndex]
            var yKeyframe = scaleValues[1][keyIndex]
    
            // 0 => XPos (frame)
            // 1 => YPos (scale value)
            // 2 => lTan
            // 3 => rTan
            var posX = xKeyframe[0] // Same in both arrays
            var xValue = xKeyframe[1]*100
            var yValue = yKeyframe[1]*100
    
            var incidenceXLeft = 0.1
            var incidenceXRight = 0.1
            var incidenceYLeft = 0.1
            var incidenceYRight = 0.1
    
            var speedRTanX = 0 
            var speedRTanY = 0 
            var speedLTanX = 0 
            var speedLTanY = 0 
            if( keyIndex > 0 )
            {
                var keyframePreviousX = scaleValues[0][keyIndex - 1]
                var tanPosXX = posX + xKeyframe[2][0] //lTan
                var tanPosXY = posX + yKeyframe[2][0] //lTan
                var posLeft = keyframePreviousX[0] //posX, same for X and Y
                incidenceXLeft = (1-((tanPosXX - posLeft) / (posX - posLeft))) * 100
                incidenceYLeft = (1-((tanPosXY - posLeft) / (posX - posLeft))) * 100
    
                speedLTanX = framerate * xKeyframe[2][1] * 100 / xKeyframe[2][0] // scale is percentage, so we * 100
                speedLTanY = framerate * yKeyframe[2][1] * 100 / yKeyframe[2][0] // scale is percentage, so we * 100
                
                // Here we need to correct, otherwise we'll get a 1-0 = 1 => max tangent left
                if( xKeyframe[2][0] == 0 )
                { incidenceXLeft = 0.1 }
                if( yKeyframe[2][0] == 0 )
                { incidenceYLeft = 0.1 }
                
                if( xKeyframe[2][1] == 0 )
                { speedLTanX = 0.0 }
                if( yKeyframe[2][1] == 0 )
                { speedLTanY = 0.0 }
            }
            
            if( keyIndex < scaleValues[0].length - 1 )
            {
                var keyframeNextX = scaleValues[0][keyIndex + 1]
                var tanPosYX = posX + xKeyframe[3][0] //rTan
                var tanPosYY = posX + yKeyframe[3][0] //rTan
                var posRight = keyframeNextX[0]
                incidenceXRight = ((tanPosYX - posX) / (posRight - posX)) * 100
                incidenceYRight = ((tanPosYY - posX) / (posRight - posX)) * 100
                
                speedRTanX = framerate * xKeyframe[3][1] * 100 / xKeyframe[3][0] // scale is percentage, so we * 100
                speedRTanY = framerate * yKeyframe[3][1] * 100 / yKeyframe[3][0] // scale is percentage, so we * 100
    
                if( xKeyframe[3][1] == 0 )
                { speedRTanX = 0 }
                if( yKeyframe[3][1] == 0 )
                { speedRTanY = 0 }
            }
    
            // Vitesse est li&é aux fps, il faut trouver le vecteur directeur lié au fps, sachans qu'une vitesse == fps, représente un vecteur [-n,-n] en lTan
            // Ensuite, avec l'influence on connais le X du pt de control, mettons 50%, pour un point en 0 et un en 1, on a X = 0.5
            // Si la vitesse est de 12, le fps de 24, on a un ratio de 2 : 24 = 12*2, donc le vecteur sera de la forme [-n*2,-n]
            // Si on a une influence de 50%, on devra donc etre a  0.5* deltaT entre 2 pts, donc 0.5 entre le point 0 et 1
            // On aura donc un vecteur de type [-0.5, -0.25], qui correspond bien a la forme [ -2n, -n ]
            // Pour 2 pt entre 10 et 20, on aurai pour 50% un X = 15, donc un vecteur [ -15, -(15/2) ]
            // Ensuite on applique ce vecteur au point pt2 [ 1, 1 ] : pt2 + vecteur = [ 1-0.5, 1-0.25 ] -> [ 0.5, 0.75 ]
            // [ 0.5, 0.75 ] sont les coords du pt de controle 
    
            // RTAN [ n, xn ]
            // rTanX / rTanY = ratio
            // speed = ratio * framerate
            // influence is good already
    
    
            var easeInX  = new KeyframeEase( speedLTanX, Math.min( Math.max( incidenceXLeft, 0.1 ), 100 ) )
            var easeOutX = new KeyframeEase( speedRTanX, Math.min( Math.max( incidenceXRight, 0.1 ), 100 ) )
            var easeInY  = new KeyframeEase( speedLTanY, Math.min( Math.max( incidenceYLeft, 0.1 ), 100 ) )
            var easeOutY = new KeyframeEase( speedRTanY, Math.min( Math.max( incidenceYRight, 0.1 ), 100 ) )
            
            var keyFrameIndex = transfoNullObject.transform.scale.addKey( posX/framerate )
            transfoNullObject.transform.scale.setValueAtKey( keyFrameIndex, [ xValue, yValue, 1 ] )
            transfoNullObject.transform.scale.setInterpolationTypeAtKey( keyFrameIndex, KeyframeInterpolationType.BEZIER, 
                                                                                        KeyframeInterpolationType.BEZIER )
            transfoNullObject.transform.scale.setTemporalEaseAtKey( keyFrameIndex, easeIn=[easeInX,easeInY,easeInX], 
                                                                                   easeOut=[easeOutX,easeOutY,easeOutX] )
        }
    }


    // Rotation
    if ( curvesJSON[4]["keys"].length > 0 )
    {
        for( keyIndex=0; keyIndex < curvesJSON[4]["keys"].length; keyIndex++ )
        {
            var keyframeJSON = curvesJSON[4]["keys"][keyIndex]
            var lTan = keyframeJSON["lTangent"]
            var rTan = keyframeJSON["rTangent"]
            var posX = keyframeJSON["time"]
            var posY = RadToDeg( keyframeJSON["value"] )

            lTan[1] = RadToDeg(lTan[1]) // rotation is in degree, so we convert
            rTan[1] = RadToDeg(rTan[1]) // rotation is in degree, so we convert
            BuildKeyFrame( transfoNullObject.transform.rotation, framerate, curvesJSON[4]["keys"], keyIndex, posX, posY, lTan, rTan, pivotX, pivotY )
        }
    }
}


function RadToDeg( value )
{
    return  value * 180 / 3.14159265359
}

function BuildScale( jsonScaleX, jsonScaleY )
{
    if( jsonScaleX.length == 0 && jsonScaleY.length == 0 )
    { return  null }

    var scaleXArray = []
    for( keyIndex=0; keyIndex < jsonScaleX.length; keyIndex++ )
    {
        var keyframeJSON = jsonScaleX[keyIndex]
        var lTan = keyframeJSON["lTangent"]
        var rTan = keyframeJSON["rTangent"]
        var posX = keyframeJSON["time"]
        var posY = keyframeJSON["value"]

        scaleXArray.push( [posX, posY, lTan, rTan] )
    }

    var scaleYArray = []
    for( keyIndex=0; keyIndex < jsonScaleY.length; keyIndex++ )
    {
        var keyframeJSON = jsonScaleY[keyIndex]
        var lTan = keyframeJSON["lTangent"]
        var rTan = keyframeJSON["rTangent"]
        var posX = keyframeJSON["time"]
        var posY = keyframeJSON["value"]

        scaleYArray.push( [posX, posY, lTan, rTan] )
    }

    if( scaleXArray.length == 0 )
    {
        FillAIntoBFrom( scaleYArray, scaleXArray, 0, 1 )
    }
    else if( scaleYArray.length == 0 )
    {
        FillAIntoBFrom( scaleXArray, scaleYArray, 0, 1 )
    }
    else
    {
        var index = 0;
        while( index < scaleXArray.length && index < scaleYArray.length )
        {
            var timeScaleX = scaleXArray[index][0]
            var timeScaleY = scaleYArray[index][0]

            if( timeScaleX != timeScaleY )
            {
                if( timeScaleX < timeScaleY )
                {
                    var lerpValue = GetInsertedKeyframeValue( scaleYArray, index, timeScaleX )
                    scaleYArray.splice( index, 0, [timeScaleX, lerpValue, [0,0], [0,0]] )
                }
                else
                {
                    var lerpValue = GetInsertedKeyframeValue( scaleXArray, index, timeScaleY )
                    scaleXArray.splice( index, 0, [timeScaleY, lerpValue, [0,0], [0,0]] )
                }
            }
            
            index++
        }

        if( scaleXArray.length < scaleYArray.length )
        {
            FillAIntoBFrom( scaleYArray, scaleXArray, index, scaleXArray[scaleXArray.length-1][1] )
        }
        else if( scaleXArray.length > scaleYArray.length )
        {
            FillAIntoBFrom( scaleXArray, scaleYArray, index, scaleYArray[scaleYArray.length-1][1] )
        }    
    }

    return  [scaleXArray, scaleYArray]
}


function GetInsertedKeyframeValue( array, index, t )
{
    if( index == 0 )
    {
        return  array[ 0 ][ 1 ]
    }
    else if( index > array.length - 1 )
    {
        return  array[ array.length - 1 ][ 1 ]
    }
    else
    {
        var previousValue = array[ index - 1][ 1 ]
        var nextValue = array[ index ][ 1 ] // Future next value, as we'll insert at index

        var prevTime = array[ index - 1][ 0 ]
        var ratio = (t - prevTime) / (array[ index ][ 0 ] - prevTime)
        return  previousValue * (1-ratio) + nextValue * ratio
    }
}

function FillAIntoBFrom( arrayA, arrayB, index, yValue )
{
    for( i=index; i<arrayA.length; i++)
    {
        arrayB.push( [arrayA[i][0], yValue, [0,0], [0,0]] )
    }
}


function BuildKeyFrame( nullObjectParameter, framerate, keysJSON, keyIndex, keyValueX, keyValueY, keyLTan, keyRTan, pivotX, pivotY )
{
    var keyFrameIndex = nullObjectParameter.addKey( keyValueX/framerate )
    nullObjectParameter.setValueAtKey( keyFrameIndex, keyValueY )
    nullObjectParameter.setInterpolationTypeAtKey( keyFrameIndex, KeyframeInterpolationType.BEZIER, KeyframeInterpolationType.BEZIER )

    var incidenceLeft = 0.1
    var incidenceRight = 0.1
    if( keyIndex > 0 )
    {
        var keyframePreviousJSON = keysJSON[keyIndex - 1]
        var tanPosX = keyValueX + keyLTan[0]
        var posLeftX = keyframePreviousJSON["time"]
        incidenceLeft = (1-((tanPosX - posLeftX) / (keyValueX - posLeftX))) * 100
    }
    
    if( keyIndex < keysJSON.length - 1 )
    {
        var keyframeNextJSON = keysJSON[keyIndex + 1]
        var tanPosX = keyValueX + keyRTan[0]
        var posRightX = keyframeNextJSON["time"]
        incidenceRight = ((tanPosX - keyValueX) / (posRightX - keyValueX)) * 100
    }

    var speedLTanX = (keyLTan[0] == 0 || keyLTan[1] == 0) ? 1 : framerate * keyLTan[1] / keyLTan[0] 
    var speedRTanX = (keyRTan[0] == 0 || keyRTan[1] == 0) ? 1 : framerate * keyRTan[1] / keyRTan[0] 
    
    incidenceLeft = incidenceLeft == 0 ? 0.1 : incidenceLeft
    incidenceRight = incidenceRight == 0 ? 0.1 : incidenceRight

    var easeIn = new KeyframeEase( speedLTanX, incidenceLeft )
    var easeOut = new KeyframeEase( speedRTanX, incidenceRight )
    nullObjectParameter.setTemporalEaseAtKey( keyFrameIndex, [easeIn], [easeOut] )
}















